package sig;

import sig.engine.AnimatedSprite;
import sig.engine.Color;
import sig.engine.Font;
import sig.engine.Key;
import sig.engine.Panel;
import sig.engine.Sprite;

import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.List;

class Player{
	AnimatedSprite spr=new AnimatedSprite("player.png",32,32);
	double x=200,y=200;
	int TAIL_SPACING=32;

	List<Tail> tail = new ArrayList<Tail>(); //The tail container
	List<PositionMarker> path = new ArrayList<PositionMarker>(); //The invisible trail container
}

class Tail{
	double x,y;
	Tail(List<PositionMarker> tailTrail,int TAIL_SPACING){
		for (int i=0;i<TAIL_SPACING;i++) {
			tailTrail.add(new PositionMarker()); //When a tail is created, we create TAIL_SPACING number of trail markers as well.
		}
	}
}

class PositionMarker{
	double x,y;
}

public class JavaProjectTemplate {
	public static final String PROGRAM_NAME="Sig's Java Project Template";
	public static int WINDOW_WIDTH=1280;
	public static int WINDOW_HEIGHT=720;
	public static Panel game;

	Player pl = new Player();
	Sprite bookSpr = new Sprite("book.png");

	JavaProjectTemplate(){
		//Initialize your game here.
	}

	void GeneratePath(){
		double prev_tail_x=pl.x;
		double prev_tail_y=pl.y;
		for (int i=0;i<pl.path.size();i++) {
			PositionMarker t = pl.path.get(i);
			double prevpos_x=t.x;
			double prevpos_y=t.y;
			t.x=prev_tail_x;
			t.y=prev_tail_y;
			prev_tail_x=prevpos_x;
			prev_tail_y=prevpos_y;
		} //Generates a very dense invisible trail.
		for (int i=0;i<pl.tail.size();i++) {
			Tail t = pl.tail.get(i);
			t.x=pl.path.get(i*pl.TAIL_SPACING).x;
			t.y=pl.path.get(i*pl.TAIL_SPACING).y; //Each tail will get placed along the index of an invisible trail based on its position in the tail container.
		}
	}

	public void updateGame(double fElapsedTime) {
		//Put game update logic in here.

		//Allow player movement using WASD keys.
		if (Key.isHeld(KeyEvent.VK_A)) {
			pl.x-=200*fElapsedTime;
		}
		if (Key.isHeld(KeyEvent.VK_D)) {
			pl.x+=200*fElapsedTime;
		}
		if (Key.isHeld(KeyEvent.VK_W)) {
			pl.y-=200*fElapsedTime;
		}
		if (Key.isHeld(KeyEvent.VK_S)) {
			pl.y+=200*fElapsedTime;
		}

		if (Key.isPressed(KeyEvent.VK_SPACE)) {
			pl.tail.add(new Tail(pl.path,pl.TAIL_SPACING));
		}
		GeneratePath();
		
	}

	public void drawGame() {
		//Put rendering logic in here.

		game.Clear(Color.BRIGHT_BLUE); //Clear the screen so all pixels from the previous frame are removed.
		game.Fill_Rect(pl.x, pl.y, 64, 64, Color.BLACK);

		for (int i=0;i<pl.tail.size();i++) {
			Tail t = pl.tail.get(i);
			game.FillCircle(t.x+22, t.y+22, 10, Color.BLUE);
			game.Draw_Circle((int)t.x+22, (int)t.y+22, 10, Color.BLACK); //Draw our tail at the proper location.
		}

		game.Draw_Text_Ext(4, 4, "SPACE to create more tail", Font.PROFONT_36, Color.RED);
	}

	public static void main(String[] args) {
		JavaProjectTemplate gameInstance=new JavaProjectTemplate();
		Panel.InitializeEngine(gameInstance);
	}
}
